@@ -99,7 +99,7 @@ union int_object {object o; fixnum i;};
 	OBJect NULL value.
 	It should not coincide with any legal object value.
 */
-#define	OBJNULL		((object)NULL)
+/* #define	OBJNULL		((object)NULL) */
 
 /*
 	Definition of each implementation type.
@@ -123,7 +123,7 @@ struct fixnum_struct {
 #define       fix_imm_fixnum(a_)        ((fixnum)a_)
 #define      mark_imm_fixnum(a_)        ((a_)=((object)((fixnum)(a_)+(LOW_IM_FIX<<1))))
 #define    unmark_imm_fixnum(a_)        ((a_)=((object)((fixnum)(a_)-(LOW_IM_FIX<<1))))
-#define        is_imm_fixnum(a_)        ((fixnum)(a_)<data_start)
+#define        is_imm_fixnum(a_)        ((fixnum)(a_)<(fixnum)OBJNULL)
 #define is_unmrkd_imm_fixnum(a_)        ((fixnum)(a_)<LOW_IM_FIX)
 #define is_marked_imm_fixnum(a_)        (is_imm_fixnum(a_)*!is_unmrkd_imm_fixnum(a_))
 #define           is_imm_fix(a_)        INT_IN_BITS(a_,LOW_SHFT-1)
@@ -158,30 +158,74 @@ struct fixnum_struct {
 #define set_fix(a_,b_)   ((a_)->FIX.FIXVAL=(b_))
 
 #define Zcdr(a_)                 (*(object *)(a_))/* ((a_)->c.c_cdr) */ /*FIXME*/
-#define is_marked(a_)            (is_imm_fixnum(Zcdr(a_)) ? is_marked_imm_fixnum(Zcdr(a_)) : (a_)->d.m)
-#define is_marked_or_free(a_)    (is_imm_fixnum(Zcdr(a_)) ? is_marked_imm_fixnum(Zcdr(a_)) : (a_)->md.mf)
-#define mark(a_)                 if (is_imm_fixnum(Zcdr(a_))) mark_imm_fixnum(Zcdr(a_)); else (a_)->d.m=1
-#define unmark(a_)               if (is_imm_fixnum(Zcdr(a_))) unmark_imm_fixnum(Zcdr(a_)); else (a_)->d.m=0
-#define is_free(a_)              (!is_imm_fixnum(a_) && !is_imm_fixnum(Zcdr(a_)) && (a_)->d.f)
-#define make_free(a_)            ({(a_)->fw=0;(a_)->d.f=1;})/*set_type_of(a_,t_other)*/
+
+#ifndef WIDE_CONS
+
+#ifndef USE_SAFE_CDR
+#define SAFE_CDR(a_)             a_
+#define imcdr(a_)                is_imm_fixnum(Zcdr(a_))
+#else
+#define SAFE_CDR(a_)             ({object _a=(a_);is_imm_fixnum(_a) ? make_fixnum1(fix(_a)) : _a;})
+#ifdef DEBUG_SAFE_CDR
+#define imcdr(a_)                (is_imm_fixnum(Zcdr(a_)) && (error("imfix cdr"),1))
+#else
+#define imcdr(a_)                0
+#endif
+#endif
+
+#else
+
+#define SAFE_CDR(a_)             a_
+#define imcdr(a_)                0
+
+#endif
+
+#define is_marked(a_)            (imcdr(a_) ? is_marked_imm_fixnum(Zcdr(a_)) : (a_)->d.m)
+#define is_marked_or_free(a_)    (imcdr(a_) ? is_marked_imm_fixnum(Zcdr(a_)) : (a_)->md.mf)
+#define mark(a_)                 if (imcdr(a_)) mark_imm_fixnum(Zcdr(a_)); else (a_)->d.m=1
+#define unmark(a_)               if (imcdr(a_)) unmark_imm_fixnum(Zcdr(a_)); else (a_)->d.m=0
+#define is_free(a_)              (!is_imm_fixnum(a_) && !imcdr(a_) && (a_)->d.f)
+#define make_free(a_)            ({(a_)->fw=0;(a_)->d.f=1;(a_)->fw|=(fixnum)OBJNULL;})/*set_type_of(a_,t_other)*/
 #define make_unfree(a_)          {(a_)->d.f=0;}
 
-#define valid_cdr(a_)            (!(a_)->d.e || is_imm_fixnum(Zcdr(a_)))
-
+#ifdef WIDE_CONS
+#define valid_cdr(a_)            0
+#else
+#define valid_cdr(a_)            (!(a_)->d.e || imcdr(a_))
+#endif
 
 #define type_of(x)       ({register object _z=(object)(x);\
                            (is_imm_fixnum(_z) ? t_fixnum : \
 			    (valid_cdr(_z) ?  (_z==Cnil ? t_symbol : t_cons)  : _z->d.t));})
+
+#ifdef WIDE_CONS
+#define TYPEWORD_TYPE_P(y_) 1
+#else
+#define TYPEWORD_TYPE_P(y_) (y_!=t_cons)
+#endif
   
-#define set_type_of(x,y) ({object _x=(object)(x);enum type _y=(y);_x->d.f=0;if (_y!=t_cons) {_x->d.e=1;_x->d.t=_y;}})
+/*Note preserve sgc flag here                                         VVV*/
+#define set_type_of(x,y) ({object _x=(object)(x);enum type _y=(y);_x->d.f=0;\
+    if (TYPEWORD_TYPE_P(_y)) {_x->d.e=1;_x->d.t=_y;_x->fw|=(fixnum)OBJNULL;}})
 
+#ifndef WIDE_CONS
 
+#define cdr_listp(x)     valid_cdr(x)
 #define consp(x)         ({register object _z=(object)(x);\
-                           (_z!=Cnil && !is_imm_fixnum(_z) && valid_cdr(_z));})
+                           (!is_imm_fixnum(_z) && valid_cdr(_z) && _z!=Cnil);})
 #define listp(x)         ({register object _z=(object)(x);\
                            (!is_imm_fixnum(_z) && valid_cdr(_z));})
 #define atom(x)          ({register object _z=(object)(x);\
-                           (_z==Cnil || is_imm_fixnum(_z) || !valid_cdr(_z));})
+                           (is_imm_fixnum(_z) || !valid_cdr(_z) || _z==Cnil);})
+
+#else
+
+#define cdr_listp(x)     listp(x)
+#define consp(x)         (type_of(x)==t_cons)
+#define listp(x)         ({object _x=x;type_of(_x)==t_cons || _x==Cnil;})
+#define atom(x)          !consp(x)
+
+#endif
 
 /* #define eql_is_eq(a_)    (is_imm_fixnum(a_) || ({enum type _tp=type_of(a_); _tp == t_cons || _tp > t_complex;})) */
 /* #define equal_is_eq(a_)  (is_imm_fixnum(a_) || type_of(a_)>t_bitvector) */
@@ -342,10 +386,16 @@ struct package {
 */
 EXTER struct package *pack_pointer;	/*  package pointer  */
 
+#ifdef WIDE_CONS
+#define Scdr(a_) (a_)->c.c_cdr
+#else
 #define Scdr(a_) ({union lispunion _t={.vw=(a_)->c.c_cdr};unmark(&_t);_t.vw;})
+#endif
 
 struct cons {
-  /* FIRSTWORD; */
+#ifdef WIDE_CONS
+  FIRSTWORD;
+#endif
   object	c_cdr;		/*  cdr  */
   object	c_car;		/*  car  */
 };
@@ -638,8 +688,11 @@ enum gcl_sm_flags {
 #endif
 
 struct random {
+
   FIRSTWORD;
-  unsigned	rnd_value;	/*  random state value  */
+
+  __gmp_randstate_struct  rnd_state;
+
 };
 
 enum chattrib {			/*  character attribute  */
@@ -858,32 +911,29 @@ struct freelist {
 	Storage manager for each type.
 */
 struct typemanager {
-	enum type
-		tm_type;	/*  type  */
-	short	tm_size;	/*  element size in bytes  */
-	short   tm_nppage;	/*  number per page  */
-	object	tm_free;	/*  free list  */
-				/*  Note that it is of type object.  */
-	long	tm_nfree;	/*  number of free elements  */
-	long	tm_npage;	/*  number of pages  */
-	long	tm_maxpage;	/*  maximum number of pages  */
-	char	*tm_name;	/*  type name  */
-	int	tm_gbccount;	/*  GBC count  */
-	object  tm_alt_free;    /*  Alternate free list (swap with tm_free) */
-	long    tm_alt_nfree;   /*  Alternate nfree (length of nfree) */
-	long	tm_alt_npage;	/*  number of pages  */
-	short   tm_sgc;         /*  this type has at least this many
-				    sgc pages */
-	short   tm_sgc_minfree;   /* number free on a page to qualify for
-				    being an sgc page */
-	short   tm_sgc_max;     /* max on sgc pages */
-	short   tm_min_grow;    /* min amount to grow when growing */
-	short   tm_max_grow;    /* max amount to grow when growing */
-	short   tm_growth_percent;  /* percent to increase maxpages */
-	short   tm_percent_free;  /* percent which must be free after a gc for this type */
-        short   tm_distinct;       /* pages of this type are distinct */
-        float   tm_adjgbccnt;
-        long    tm_opt_maxpage;
+  enum type tm_type;             /*  type  */
+  long	    tm_size;             /*  element size in bytes  */
+  long      tm_nppage;           /*  number per page  */
+  object    tm_free;             /*  free list  */
+				 /*  Note that it is of type object.  */
+  long	    tm_nfree;            /*  number of free elements  */
+  long	    tm_npage;            /*  number of pages  */
+  long	    tm_maxpage;          /*  maximum number of pages  */
+  char	   *tm_name;             /*  type name  */
+  long	    tm_gbccount;         /*  GBC count  */
+  object    tm_alt_free;         /*  Alternate free list (swap with tm_free) */
+  long      tm_alt_nfree;        /*  Alternate nfree (length of nfree) */
+  long	    tm_alt_npage;        /*  number of pages  */
+  long      tm_sgc;              /*  this type has at least this many sgc pages */
+  long      tm_sgc_minfree;      /*  number free on a page to qualify for being an sgc page */
+  long      tm_sgc_max;          /* max on sgc pages */
+  long      tm_min_grow;         /* min amount to grow when growing */
+  long      tm_max_grow;         /* max amount to grow when growing */
+  long      tm_growth_percent;   /* percent to increase maxpages */
+  long      tm_percent_free;     /* percent which must be free after a gc for this type */
+  long      tm_distinct;         /* pages of this type are distinct */
+  float     tm_adjgbccnt;
+  long      tm_opt_maxpage;
 };
 
 
@@ -897,6 +947,7 @@ EXTER struct typemanager tm_table[ 32  /* (int) t_relocatable */];
 /*
 	Contiguous block header.
 */
+EXTER bool prefer_low_mem_contblock;
 struct contblock {		/*  contiguous block header  */
 	int	cb_size;	/*  size in bytes  */
 	struct contblock
@@ -917,18 +968,16 @@ EXTER struct contblock *old_cb_pointer;	/*  old contblock pointer when in SGC  *
 	Variables for memory management.
 */
 EXTER long ncb;			/*  number of contblocks  */
-/* int ncbpage;			  number of contblock pages  */
 #define ncbpage tm_table[t_contiguous].tm_npage
 #define maxcbpage tm_table[t_contiguous].tm_maxpage
 #define cbgbccount tm_table[t_contiguous].tm_gbccount  
   
 
-/* int maxcbpage; maximum number of contblock pages  */
-EXTER 
-long holepage;			/*  hole pages  */
+EXTER long holepage;			/*  hole pages  */
 #define nrbpage tm_table[t_relocatable].tm_npage
+#define maxrbpage tm_table[t_relocatable].tm_maxpage
 #define rbgbccount tm_table[t_relocatable].tm_gbccount
-/* int nrbpage;			  number of relblock pages  */
+EXTER long new_holepage,starting_hole_div,starting_relb_heap_mult;
   
 
 #ifdef SGC
@@ -1147,3 +1196,9 @@ EXTER struct symbol Dotnil_body;
 #define endp(a) endp_prop(a)
     
 #define proper_list(a) (type_of(a)==t_cons || (a)==Cnil)
+
+#define IMMNIL(x) (is_imm_fixnum(x)||x==Cnil)
+
+#define eql(a_,b_)    ({register object _a=(a_);register object _b=(b_);_a==_b || (!IMMNIL(_a)&&!IMMNIL(_b)&&eql1(_a,_b));})
+#define equal(a_,b_)  ({register object _a=(a_);register object _b=(b_);_a==_b || (!IMMNIL(_a)&&!IMMNIL(_b)&&equal1(_a,_b));})
+#define equalp(a_,b_) ({register object _a=(a_);register object _b=(b_);_a==_b || (_a!=Cnil&&_b!=Cnil&&equalp1(_a,_b));})
