@@ -61,6 +61,8 @@ sbrk1(n)
 #define sbrk sbrk1
 #endif /* DEBUG_SBRK */
 
+long starting_hole_div=10;
+long starting_relb_heap_mult=2;
 long new_holepage;
 long resv_pages=40;
 
@@ -90,14 +92,16 @@ add_page_to_contblock_list(void *p,fixnum m) {
     contblock_list_tail=p;
   }
   
+  bzero(pagetochar(page(pp)),CB_DATA_START(pp)-(void *)pagetochar(page(pp)));
 #ifdef SGC
-  if (sgc_enabled)
+  if (sgc_enabled && tm_table[t_contiguous].tm_sgc) {
+    memset(CB_SGCF_START(pp),-1,CB_DATA_START(pp)-CB_SGCF_START(pp));
     pp->sgc_flags=SGC_PAGE_FLAG;
+    }
 #endif
   
   ncbpage+=m;
-  p+=mbytes(m)+sizeof(struct pageinfo);
-  insert_contblock(p,PAGESIZE*m-(p-(void *)pp));
+  insert_contblock(CB_DATA_START(pp),CB_DATA_END(pp)-CB_DATA_START(pp));
 
 }
 
@@ -269,6 +273,7 @@ set_tm_maxpage(struct typemanager *tm,fixnum n) {
   
   fixnum r=tm->tm_type==t_relocatable,j=tm->tm_maxpage,z=(n-j)*(r ? 2 : 1);
   if (z>available_pages) return 0;
+  if (r && 2*n+page(REAL_RB_START)>real_maxpage) return 0;
   available_pages-=z;
   tm->tm_adjgbccnt*=((double)j)/n;
   tm->tm_maxpage=n;
@@ -301,14 +306,14 @@ add_page_to_freelist(char *p, struct typemanager *tm) {
  }
 
  x= (object)pagetochar(page(p));
- set_type_of(x,t);
+ /* set_type_of(x,t); */
  make_free(x);
 
 #ifdef SGC
 
  if (sgc_enabled && tm->tm_sgc)
    pp->sgc_flags=SGC_PAGE_FLAG;
- if (pp->type!=t_cons)
+ if (TYPEWORD_TYPE_P(pp->type))
    x->d.s=(sgc_enabled && tm->tm_sgc) ? SGC_RECENT : SGC_NORMAL;
 
  /* array headers must be always writable, since a write to the
@@ -398,7 +403,7 @@ grow_linear(fixnum old, fixnum fract, fixnum grow_min, fixnum grow_max,fixnum ma
 
    20040403 CM */
 
-DEFVAR("*OPTIMIZE-MAXIMUM-PAGES*",sSAoptimize_maximum_pagesA,SI,sLt,"");
+DEFVAR("*OPTIMIZE-MAXIMUM-PAGES*",sSAoptimize_maximum_pagesA,SI,sLnil,"");
 #define OPTIMIZE_MAX_PAGES (sSAoptimize_maximum_pagesA ==0 || sSAoptimize_maximum_pagesA->s.s_dbind !=sLnil) 
 DEFVAR("*NOTIFY-OPTIMIZE-MAXIMUM-PAGES*",sSAnotify_optimize_maximum_pagesA,SI,sLnil,"");
 #define MMAX_PG(a_) (a_)->tm_maxpage
@@ -476,25 +481,37 @@ Use ALLOCATE to expand the space.",
 #else
 #define TOTAL_THIS_TYPE(tm) (tm->tm_nppage * tm->tm_npage)
 #endif
+bool prefer_low_mem_contblock=FALSE;
+
 inline void *
 alloc_from_freelist(struct typemanager *tm,fixnum n) {
 
-  void *p;
+  void *p,*v,*vp;
   struct contblock **cbpp;
   fixnum i;
 
   switch (tm->tm_type) {
 
   case t_contiguous:
-    for(cbpp= &cb_pointer; (*cbpp)!=NULL; cbpp= &(*cbpp)->cb_link)
+    for (cbpp= &cb_pointer,v=(void *)-1,vp=NULL; (*cbpp)!=NULL; cbpp= &(*cbpp)->cb_link)
       if ((*cbpp)->cb_size >= n) {
-	p=(void *)(*cbpp);
-	i=(*cbpp)->cb_size-n;
-	*cbpp=(*cbpp)->cb_link;
-	--ncb;
-	insert_contblock(p+n,i);
-	return(p);
+	if (!prefer_low_mem_contblock) {
+	  vp=cbpp;
+	  break;
+	} else if ((void *)(*cbpp)<v) {
+	  v=*cbpp;
+	  vp=cbpp;
+	}
       }
+    if (vp) {
+      cbpp=vp;
+      p=(void *)(*cbpp);
+      i=(*cbpp)->cb_size-n;
+      *cbpp=(*cbpp)->cb_link;
+      --ncb;
+      insert_contblock(p+n,i);
+      return(p);
+    }
     break;
 
   case t_relocatable:
@@ -733,6 +750,14 @@ alloc_relblock(size_t n) {
 
 }
 
+static inline void
+load_cons(object p,object a,object d) {
+#ifdef WIDE_CONS
+  set_type_of(p,t_cons);
+#endif
+  p->c.c_cdr=SAFE_CDR(d);
+  p->c.c_car=a;
+}
 
 inline object
 make_cons(object a,object d) {
@@ -740,8 +765,7 @@ make_cons(object a,object d) {
   static struct typemanager *tm=tm_table+t_cons;/*FIXME*/
   object obj=alloc_mem(tm,tm->tm_size);
 
-  obj->c.c_car = a;
-  obj->c.c_cdr = d;
+  load_cons(obj,a,d);
 
   pageinfo(obj)->in_use++;
 
@@ -751,12 +775,10 @@ make_cons(object a,object d) {
 
 
 
-inline object on_stack_cons(object x, object y)
-{object p = (object) alloca_val;
- set_type_of(p,t_cons);
- p->c.c_car=x;
- p->c.c_cdr=y;
- return p;
+inline object on_stack_cons(object x, object y) {
+  object p = (object) alloca_val;
+  load_cons(p,x,y);
+  return p;
 }
 
 
@@ -937,8 +959,8 @@ init_tm(enum type t, char *name, int elsize, int nelts, int sgc,int distinct) {
     return;
   }
   tm_table[(int)t].tm_type = t;
-  tm_table[(int)t].tm_size = ROUND_UP_PTR(elsize);
-  tm_table[(int)t].tm_nppage = (PAGESIZE-sizeof(struct pageinfo))/ROUND_UP_PTR(elsize);
+  tm_table[(int)t].tm_size = elsize ? ROUND_UP_PTR(elsize) : 1;
+  tm_table[(int)t].tm_nppage = (PAGESIZE-sizeof(struct pageinfo))/tm_table[(int)t].tm_size;
   tm_table[(int)t].tm_free = OBJNULL;
   tm_table[(int)t].tm_nfree = 0;
   /* tm_table[(int)t].tm_nused = 0; */
@@ -952,8 +974,7 @@ init_tm(enum type t, char *name, int elsize, int nelts, int sgc,int distinct) {
 #ifdef SGC	
   tm_table[(int)t].tm_sgc = sgc;
   tm_table[(int)t].tm_sgc_max = 3000;
-  tm_table[(int)t].tm_sgc_minfree = (int)
-    (0.4 * tm_table[(int)t].tm_nppage);
+  tm_table[(int)t].tm_sgc_minfree = (0.4 * tm_table[(int)t].tm_nppage);
 #endif
   
 }
@@ -1041,14 +1062,13 @@ gcl_init_alloc(void) {
   
   set_maxpage();
   
-  /* holepage = INIT_HOLEPAGE; */
-  holepage = available_pages/10;
+  holepage=new_holepage;
+
 #ifdef GCL_GPROF
   if (holepage<textpage)
      holepage=textpage;
 #endif
 
-  new_holepage = available_pages/10;
   /* Unused (at present) tm_distinct flag added.  Note that if cons
      and fixnum share page types, errors will be introduced.
 
@@ -1087,10 +1107,8 @@ gcl_init_alloc(void) {
   init_tm(t_afun, "AAFUN", sizeof(struct sfun), 0 ,1,0);
   init_tm(t_cfdata, "cCFDATA", sizeof(struct cfdata), 102 ,1,0);
   init_tm(t_spice, "!SPICE", sizeof(struct spice), 4096 ,1,0);
-  init_tm(t_relocatable, "%RELOCATABLE-BLOCKS", 1000,0,20,0);
-  init_tm(t_contiguous, "_CONTIGUOUS-BLOCKS", 1001,0,20,0);
-  tm_table[t_relocatable].tm_nppage = PAGESIZE;
-  tm_table[t_contiguous].tm_nppage = PAGESIZE;
+  init_tm(t_relocatable, "%RELOCATABLE-BLOCKS", 0,0,20,1);
+  init_tm(t_contiguous, "_CONTIGUOUS-BLOCKS", 0,0,20,1);
   
   
   ncb = 0;
@@ -1101,7 +1119,7 @@ gcl_init_alloc(void) {
     set_tm_maxpage(tm_table+t_contiguous,textpage);
 #endif
 
-  set_tm_maxpage(tm_table+t_relocatable,5);
+  set_tm_maxpage(tm_table+t_relocatable,19);
   nrbpage=0;
 
 #ifdef __linux__
@@ -1183,7 +1201,7 @@ DEFUN_NEW("ALLOCATE-SGC",object,fSallocate_sgc,SI
   x2=make_fixnum((100*tm->tm_sgc_minfree)/tm->tm_nppage);
   res= list(3,x,x1,x2);
   
-  if(min<0 || max< min || min > 3000 || free_percent < 0 || free_percent > 100)
+  if(min<0 || max< min || free_percent < 0 || free_percent > 100)
     goto END;
   tm->tm_sgc_max=max;
   tm->tm_sgc=min;
@@ -1340,6 +1358,8 @@ DEFUN_NEW("GET-HOLE-SIZE",object,fSget_hole_size,SI,0,0,NONE,OO,OO,OO,OO,(void),
 
 static unsigned long start,end,gprof_on;
 static void *initial_monstartup_pointer;
+bool capture_gprof_array=1;
+object gprof_array=Cnil;
 
 void
 gprof_cleanup(void) {
@@ -1348,6 +1368,7 @@ gprof_cleanup(void) {
 
   if (initial_monstartup_pointer) {
     _mcleanup();
+    gprof_array=Cnil;
     gprof_on=0;
   }
 
@@ -1384,6 +1405,7 @@ DEFUN_NEW("GPROF-START",object,fSgprof_start,SI
   if (!gprof_on) {
     start=start ? start : textstart;
     end=end ? end : textend;
+    capture_gprof_array=1;
     monstartup(start,end);
     gprof_on=1;
     if (!n && atexit(gprof_cleanup)) {
@@ -1423,6 +1445,7 @@ DEFUN_NEW("GPROF-QUIT",object,fSgprof_quit,SI
   if (chdir(P_tmpdir))
     FEerror("Cannot change directory to tmpdir", 0);
   _mcleanup();
+  gprof_array=Cnil;
   if (snprintf(b1,sizeof(b1),"gprof %s",kcl_self)<=0)
     FEerror("Cannot write gprof command line", 0);
   if (!(pp=popen(b1,"r")))
@@ -1442,34 +1465,23 @@ DEFUN_NEW("GPROF-QUIT",object,fSgprof_quit,SI
 
 #endif
 
-DEFUN_NEW("SET-HOLE-SIZE",object,fSset_hole_size,SI
-       ,1,2,NONE,OI,IO,OO,OO,(fixnum npages,...),"")
-{
-
-  int nargs=VFUN_NARGS;
-  int reserve;
-  va_list ap;
-
-  reserve=30;
-  if (nargs>=2) {
-    va_start(ap,npages);
-    reserve=va_arg(ap,fixnum);
-    va_end(ap);
-  }
+DEFUN_NEW("SET-STARTING-HOLE-DIVISOR",fixnum,fSset_starting_hole_divisor,SI,1,1,NONE,II,OO,OO,OO,(fixnum div),"") {
+  if (div>0 && div <100)
+    starting_hole_div=div;
+  return starting_hole_div;
+}
   
-  if (npages < 1 ||
-      npages > real_maxpage - page(heap_end)
-      - 2*nrbpage - real_maxpage/32)
-    FEerror("Illegal value for the hole size.", 0);
-  new_holepage = npages;
-  if (VFUN_NARGS ==2)
-    {
-      if (reserve <0 || reserve > new_holepage)
-	FEerror("Illegal value for the hole size.", 0);
-      reserve_pages_for_signal_handler = reserve;}
+DEFUN_NEW("SET-STARTING-RELBLOCK-HEAP-MULTIPLE",fixnum,fSset_starting_relb_heap_multiple,SI,1,1,NONE,II,OO,OO,OO,(fixnum mult),"") {
+  if (mult>=0)
+    starting_relb_heap_mult=mult;
+  return starting_relb_heap_mult;
+}
   
-  RETURN2(make_fixnum(npages),
-	  make_fixnum(reserve_pages_for_signal_handler));
+DEFUN_NEW("SET-HOLE-SIZE",object,fSset_hole_size,SI,1,2,NONE,OI,IO,OO,OO,(fixnum npages,...),"") {
+
+  printf("This function is obsolete -- use SET-STARTING-HOLE-DIVISOR instead\n");
+
+  RETURN2(make_fixnum(new_holepage),make_fixnum(reserve_pages_for_signal_handler));
 
 }
 
@@ -1601,6 +1613,10 @@ malloc(size_t size) {
 #ifdef GCL_GPROF
   if (!initflag && size>(textend-textstart) && !initial_monstartup_pointer) 
     initial_monstartup_pointer=malloc_list->c.c_car->st.st_self;
+  if (gprof_array==Cnil && capture_gprof_array && size>(textend-textstart)) {
+    gprof_array=malloc_list->c.c_car;
+    capture_gprof_array=0;
+  }
 #endif
   
   in_malloc=0;
