@@ -122,10 +122,6 @@ off_check(void *v,void *ve,fixnum i,struct pageinfo *pi) {
 #endif
 
 
-#define CB_DATA_SIZE(z_)   ({fixnum _z=(z_);_z*PAGESIZE-mbytes(_z)-sizeof(struct pageinfo);})
-#define CB_MARK_START(pi_) ((void *)(pi_)+sizeof(struct pageinfo))
-#define CB_DATA_START(pi_) ({struct pageinfo *_pi=(pi_);CB_MARK_START(_pi)+mbytes(_pi->in_use);})
-
 inline struct pageinfo *
 get_pageinfo(void *x) {
   struct pageinfo *v=contblock_list_head;void *vv;
@@ -134,8 +130,8 @@ get_pageinfo(void *x) {
 }
 
 inline char
-get_mark_bit(struct pageinfo *pi,char *x) {
-  char *v=CB_MARK_START(pi),*ve=CB_DATA_START(pi);
+get_bit(char *v,struct pageinfo *pi,void *x) {
+  void *ve=CB_DATA_START(pi);
   fixnum off=(x-ve)>>LOG_BYTES_CONTBLOCK,i=off>>LOG_BITS_CHAR,s=off&~(~0UL<<LOG_BITS_CHAR);
 #ifdef CONTBLOCK_MARK_DEBUG
   off_check(v,ve,i,pi);
@@ -144,8 +140,8 @@ get_mark_bit(struct pageinfo *pi,char *x) {
 }
 
 inline void
-set_mark_bit(struct pageinfo *pi,char *x) {
-  char *v=CB_MARK_START(pi),*ve=CB_DATA_START(pi);
+set_bit(char *v,struct pageinfo *pi,void *x) {
+  void *ve=CB_DATA_START(pi);
   fixnum off=(x-ve)>>LOG_BYTES_CONTBLOCK,i=off>>LOG_BITS_CHAR,s=off&~(~0UL<<LOG_BITS_CHAR);
 #ifdef CONTBLOCK_MARK_DEBUG
   off_check(v,ve,i,pi);
@@ -153,6 +149,99 @@ set_mark_bit(struct pageinfo *pi,char *x) {
   v[i]|=(1UL<<s);
 }
 
+#define bit_get(v,i,s) ((v[i]>>s)&0x1)
+#define bit_set(v,i,s) (v[i]|=(1UL<<s))
+#define ptr_get(v,i,s) (v+(((i<<LOG_BITS_CHAR)|s)<<LOG_BYTES_CONTBLOCK))
+#define ptr_set(x,v,i,s) ({fixnum _o=(x-v)>>LOG_BYTES_CONTBLOCK;i=_o>>LOG_BITS_CHAR;s=_o&~(~0UL<<LOG_BITS_CHAR);})
+
+inline void
+set_bits(char *v,struct pageinfo *pi,void *x1,void *x2) {
+
+  void *ds=CB_DATA_START(pi);
+  fixnum i1,s1,i2,s2,se;
+
+  ptr_set(x1,ds,i1,s1);
+  ptr_set(x2,ds,i2,s2);
+
+  if (i1==i2) {
+    se=s2;
+    s2=0;
+  } else
+    se=CHAR_SIZE;
+
+  for (;s1<se;s1++)
+    bit_set(v,i1,s1);
+  if (i2>++i1) memset(v+i1,-1,(i2-i1));
+  for (;--s2>=0;)
+    bit_set(v,i2,s2);
+
+}
+
+inline void *
+get_bits(char *v,struct pageinfo *pi,void *x) {
+
+  void *ds=CB_DATA_START(pi),*de=CB_DATA_END(pi);
+  fixnum i,s,ie=mbytes(pi->in_use);
+  bool z;
+  char cz;
+
+  ptr_set(x,ds,i,s);
+  z=bit_get(v,i,s);
+  cz=z?-1:0;
+
+  for (;++s<CHAR_SIZE && z==bit_get(v,i,s););
+  if (s==CHAR_SIZE) {
+    for (;++i<ie && v[i]==cz;);
+    if (i<ie) for (s=-1;++s<CHAR_SIZE && z==bit_get(v,i,s);); else s=CHAR_SIZE-1;
+    /* massert(s<CHAR_SIZE); */
+  }
+  ds=ptr_get(ds,i,s);
+
+  /* for (;x<ds;x+=sizeof(struct contblock)) */
+  /*   massert(z==get_bit(v,pi,x)); */
+
+  return ds<de ? ds : de;
+}
+
+inline char
+get_mark_bit(struct pageinfo *pi,void *x) {
+  return get_bit(CB_MARK_START(pi),pi,x);
+}
+
+inline void
+set_mark_bit(struct pageinfo *pi,void *x) {
+  set_bit(CB_MARK_START(pi),pi,x);
+}
+
+inline void *
+get_mark_bits(struct pageinfo *pi,void *x) {
+  return get_bits(CB_MARK_START(pi),pi,x);
+}
+
+inline void
+set_mark_bits(struct pageinfo *pi,void *x1,void *x2) {
+  set_bits(CB_MARK_START(pi),pi,x1,x2);
+}
+
+inline char
+get_sgc_bit(struct pageinfo *pi,void *x) {
+  return get_bit(CB_SGCF_START(pi),pi,x);
+}
+
+inline void
+set_sgc_bit(struct pageinfo *pi,void *x) {
+  set_bit(CB_SGCF_START(pi),pi,x);
+}
+
+inline void *
+get_sgc_bits(struct pageinfo *pi,void *x) {
+  return get_bits(CB_SGCF_START(pi),pi,x);
+}
+
+inline void
+set_sgc_bits(struct pageinfo *pi,void *x1,void *x2) {
+  set_bits(CB_SGCF_START(pi),pi,x1,x2);
+}
 
 #ifdef KCLOVM
 void mark_all_stacks();
@@ -212,19 +301,17 @@ enter_mark_origin(object *p) {
 
 }
 
-static void
+inline void
 mark_cons(object x) {
   
-  object d;
-
   do {
-    d=x->c.c_cdr;
+    object d=x->c.c_cdr;
     mark(x);
     mark_object(x->c.c_car);
     x=d;
-    if (NULL_OR_ON_C_STACK(x) || is_marked_or_free(x))
+    if (NULL_OR_ON_C_STACK(x) || is_marked_or_free(x))/*catches Cnil*/
       return;
-  } while (type_of(x)==t_cons);
+  } while (cdr_listp(x));
   mark_object(x);
 
 }
@@ -237,18 +324,26 @@ mark_cons(object x) {
 #define MARK_LINK_ARRAY(x_) ((*(unsigned long *)(x_))|=1UL)
 #define CLEAR_LINK_ARRAY(x_) ((*(unsigned long *)(x_))&=~(1UL))
 
+/* #define COLLECT_RELBLOCK_P (what_to_collect == t_relocatable || what_to_collect == t_contiguous) */
+bool collect_both=0;
+
+#define COLLECT_RELBLOCK_P (what_to_collect == t_relocatable || collect_both)
+
 static void
 mark_link_array(void *v,void *ve) {
 
   void **p,**pe;
 
+  if (NULL_OR_ON_C_STACK(v))
+    return;
+
   if (sLAlink_arrayA->s.s_dbind==Cnil)
     return;
 
   p=(void *)sLAlink_arrayA->s.s_dbind->v.v_self;
   pe=(void *)p+sLAlink_arrayA->s.s_dbind->v.v_fillp;
 
-  if (is_marked(sLAlink_arrayA->s.s_dbind)
+  if (is_marked(sLAlink_arrayA->s.s_dbind) && COLLECT_RELBLOCK_P
 #ifdef SGC
       && (!sgc_enabled || SGC_RELBLOCK_P(sLAlink_arrayA->s.s_dbind->v.v_self))
 #endif
@@ -317,7 +412,6 @@ sweep_link_array(void) {
 
 }
 
-
 static void
 mark_object(object x) {
   
@@ -333,18 +427,20 @@ mark_object(object x) {
      should be dead, we dont want to mark it. -wfs
   */
   
-  if (NULL_OR_ON_C_STACK(x))
-    return;
-  if (is_marked_or_free(x))
+  if (NULL_OR_ON_C_STACK(x) || is_marked_or_free(x))
     return;
 
-  if ((tp=type_of(x))==t_cons) {
+  tp=type_of(x);
+
+  if (tp==t_cons) {
     mark_cons(x);
     return;
   }
 
   mark(x);
+
   switch (tp) {
+
   case t_fixnum:
     break;
     
@@ -373,15 +469,11 @@ mark_object(object x) {
     mark_object(x->s.s_dbind);
     if (x->s.s_self == NULL)
       break;
-    if ((int)what_to_collect >= (int)t_contiguous) {
-      if (inheap(x->s.s_self)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(x->s.s_self,
-			 x->s.s_fillp);
-      } else
-	x->s.s_self =
-	  copy_relblock(x->s.s_self, x->s.s_fillp);
-    }
+    if (inheap(x->s.s_self)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(x->s.s_self,x->s.s_fillp);
+    } else if (COLLECT_RELBLOCK_P)
+	x->s.s_self = copy_relblock(x->s.s_self, x->s.s_fillp);
     break;
     
   case t_package:
@@ -400,15 +492,6 @@ mark_object(object x) {
 		     x->p.p_external_size*sizeof(object));
     break;
     
-  case t_cons:
-    /*
-      mark_object(x->c.c_car);
-      x = x->c.c_cdr;
-      goto BEGIN;
-    */
-    mark_cons(x);
-    break;
-    
   case t_hashtable:
     mark_object(x->ht.ht_rhsize);
     mark_object(x->ht.ht_rhthresh);
@@ -418,28 +501,22 @@ mark_object(object x) {
       mark_object(x->ht.ht_self[i].hte_key);
       mark_object(x->ht.ht_self[i].hte_value);
     }
-    if ((short)what_to_collect >= (short)t_contiguous) {
-      if (inheap(x->ht.ht_self)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)x->ht.ht_self,j*sizeof(struct htent));
-      } else
-	x->ht.ht_self=(void *)copy_relblock((char *)x->ht.ht_self,j*sizeof(struct htent));;
-    }
+    if (inheap(x->ht.ht_self)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock((char *)x->ht.ht_self,j*sizeof(struct htent));
+    } else if (COLLECT_RELBLOCK_P)
+      x->ht.ht_self=(void *)copy_relblock((char *)x->ht.ht_self,j*sizeof(struct htent));;
     break;
     
   case t_array:
     if ((x->a.a_displaced) != Cnil)
       mark_displaced_field(x);
-    if ((int)what_to_collect >= (int)t_contiguous &&
-	x->a.a_dims != NULL) {
+    if (x->a.a_dims != NULL) {
       if (inheap(x->a.a_dims)) {
 	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)(x->a.a_dims),
-			 sizeof(int)*x->a.a_rank);
-      } else
-	x->a.a_dims = (int *)
-	  copy_relblock((char *)(x->a.a_dims),
-			sizeof(int)*x->a.a_rank);
+	  mark_contblock((char *)(x->a.a_dims),sizeof(int)*x->a.a_rank);
+      } else if (COLLECT_RELBLOCK_P)
+	x->a.a_dims = (int *) copy_relblock((char *)(x->a.a_dims),sizeof(int)*x->a.a_rank);
     }
     if ((enum aelttype)x->a.a_elttype == aet_ch)
       goto CASE_STRING;
@@ -463,8 +540,8 @@ mark_object(object x) {
      }}
     case aet_lf:
       j= sizeof(longfloat)*x->lfa.lfa_dim;
-      if (((int)what_to_collect >= (int)t_contiguous) &&
-	  !(inheap(cp))) ROUND_RB_POINTERS_DOUBLE;
+      if ((COLLECT_RELBLOCK_P) &&  !(inheap(cp)))
+	ROUND_RB_POINTERS_DOUBLE;/*FIXME gc space violation*/
       break;
     case aet_char:
     case aet_uchar:
@@ -494,18 +571,17 @@ mark_object(object x) {
     cp = (char *)p;
     j *= sizeof(object);
   COPY:
-    if ((int)what_to_collect >= (int)t_contiguous) {
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      } else if (x->a.a_displaced == Cnil) {
+    if (inheap(cp)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(cp, j);
+    } else if (COLLECT_RELBLOCK_P) {
+      if (x->a.a_displaced == Cnil) {
 #ifdef HAVE_ALLOCA
 	if (!NULL_OR_ON_C_STACK(cp))  /* only if body of array not on C stack */
 #endif			  
-	  x->a.a_self = (object *)copy_relblock(cp, j);}
-      else if (x->a.a_displaced->c.c_car == Cnil) {
-	i = (long)(object *)copy_relblock(cp, j)
-	  - (long)(x->a.a_self);
+	  x->a.a_self = (object *)copy_relblock(cp, j);
+      } else if (x->a.a_displaced->c.c_car == Cnil) {
+	i = (long)(object *)copy_relblock(cp, j)  - (long)(x->a.a_self);
 	adjust_displaced(x, i);
       }
     }
@@ -539,7 +615,7 @@ mark_object(object x) {
       if (inheap(cp)) {
 	if (what_to_collect == t_contiguous)
 	  mark_contblock(cp, j);
-      } else{
+      } else if (COLLECT_RELBLOCK_P) {
 	MP_SELF(x) = (void *) copy_relblock(cp, j);}}
 #endif /* not GMP_USE_MALLOC */
     break;
@@ -553,11 +629,11 @@ mark_object(object x) {
     if (cp == NULL)
       break;
   COPY_STRING:
-    if ((int)what_to_collect >= (int)t_contiguous) {
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      } else if (x->st.st_displaced == Cnil)
+    if (inheap(cp)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(cp, j);
+    } else if (COLLECT_RELBLOCK_P) {
+      if (x->st.st_displaced == Cnil)
 	x->st.st_self = copy_relblock(cp, j);
       else if (x->st.st_displaced->c.c_car == Cnil) {
 	i = copy_relblock(cp, j) - cp;
@@ -586,21 +662,18 @@ mark_object(object x) {
     p = x->str.str_self;
     if (p == NULL)
       break;
-    {object def=x->str.str_def;
-    unsigned char * s_type = &SLOT_TYPE(def,0);
-    unsigned short *s_pos= & SLOT_POS(def,0);
-    for (i = 0, j = S_DATA(def)->length;  i < j;  i++)
-      if (s_type[i]==0) mark_object(STREF(object,x,s_pos[i]));
-    if ((int)what_to_collect >= (int)t_contiguous) {
+    {
+      object def=x->str.str_def;
+      unsigned char * s_type = &SLOT_TYPE(def,0);
+      unsigned short *s_pos= & SLOT_POS(def,0);
+      for (i = 0, j = S_DATA(def)->length;  i < j;  i++)
+	if (s_type[i]==0) mark_object(STREF(object,x,s_pos[i]));
       if (inheap(x->str.str_self)) {
 	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)p,
-			 S_DATA(def)->size);
-	
-      } else
-	x->str.str_self = (object *)
-	  copy_relblock((char *)p, S_DATA(def)->size);
-    }}
+	  mark_contblock((char *)p,S_DATA(def)->size);
+      } else if (COLLECT_RELBLOCK_P)
+	x->str.str_self = (object *)copy_relblock((char *)p, S_DATA(def)->size);
+    }
     break;
     
   case t_stream:
@@ -656,7 +729,24 @@ mark_object(object x) {
     }
     break;
     
+#define MARK_CP(a_,b_) {fixnum _t=(b_);if (inheap(a_)) {\
+	if (what_to_collect == t_contiguous) mark_contblock((void *)(a_),_t); \
+      } else if (COLLECT_RELBLOCK_P) (a_)=(void *)copy_relblock((void *)(a_),_t);}
+
+#define MARK_MP(a_) {if ((a_)->_mp_d) \
+                        MARK_CP((a_)->_mp_d,(a_)->_mp_alloc*MP_LIMB_SIZE);}
+
   case t_random:
+    if ((int)what_to_collect >= (int)t_contiguous) {
+      MARK_MP(x->rnd.rnd_state._mp_seed);
+#if __GNU_MP_VERSION < 4 || (__GNU_MP_VERSION  == 4 && __GNU_MP_VERSION_MINOR < 2)
+      if (x->rnd.rnd_state._mp_algdata._mp_lc) {
+	MARK_MP(x->rnd.rnd_state._mp_algdata._mp_lc->_mp_a);
+	if (!x->rnd.rnd_state._mp_algdata._mp_lc->_mp_m2exp) MARK_MP(x->rnd.rnd_state._mp_algdata._mp_lc->_mp_m);
+	MARK_CP(x->rnd.rnd_state._mp_algdata._mp_lc,sizeof(*x->rnd.rnd_state._mp_algdata._mp_lc));
+      }
+#endif
+    }
     break;
     
   case t_readtable:
@@ -710,12 +800,7 @@ mark_object(object x) {
       {int i=x->cfd.cfd_fillp;
       while(i-- > 0)
 	mark_object(x->cfd.cfd_self[i]);}
-    if (x->cfd.cfd_start == NULL)
-      break;
     if (what_to_collect == t_contiguous) {
-      if (!MAYBE_DATA_P((x->cfd.cfd_start)) ||
-	  get_mark_bit(get_pageinfo(x->cfd.cfd_start),x->cfd.cfd_start))
-	break;
       mark_contblock(x->cfd.cfd_start, x->cfd.cfd_size);
       mark_link_array(x->cfd.cfd_start,x->cfd.cfd_start+x->cfd.cfd_size);
     }
@@ -1019,46 +1104,18 @@ mark_c_stack(jmp_buf env1, int n, void (*fn)(void *,void *,int)) {
 static void
 sweep_phase(void) {
 
-  STATIC long i, j, k;
+  STATIC long j, k;
   STATIC object x;
   STATIC char *p;
   STATIC struct typemanager *tm;
   STATIC object f;
   STATIC struct pageinfo *v;
   
-#ifdef DEBUG
-  if (debug)
-    printf("type map\n");
-#endif
   for (v=cell_list_head;v;v=v->next) {
-    enum type tp=v->type;
 
-    i=page(v);
-
-    if (tp == (int)t_contiguous) {
-      if (debug) {
-	printf("-");
-	continue;
-      }
-    }
-    if (tp >= (int)t_end)
-      continue;
+    tm = tm_of((enum type)v->type);
     
-    tm = tm_of((enum type)tp);
-    
-    /*
-      general sweeper
-    */
-    
-#ifdef DEBUG
-    if (debug) {
-      printf("%c", tm->tm_name[0]);
-      /*
-	fflush(stdout);
-      */
-    }
-#endif
-    p = pagetochar(i);
+    p = pagetochar(page(v));
     f = tm->tm_free;
     k = 0;
     for (j = tm->tm_nppage; j > 0; --j, p += tm->tm_size) {
@@ -1070,12 +1127,6 @@ sweep_phase(void) {
 	continue;
       }
 
-#ifdef GMP_USE_MALLOC
-      if (x->d.t == t_bignum) {
-	mpz_clear(MP(x));
-      }
-#endif
-
       SET_LINK(x,f);
       make_free(x);
       f = x;
@@ -1083,15 +1134,10 @@ sweep_phase(void) {
     }
     tm->tm_free = f;
     tm->tm_nfree += k;
-    pagetoinfo(i)->in_use-=k;
+    pagetoinfo(page(v))->in_use-=k;
     
   }
-#ifdef DEBUG
-  if (debug) {
-    putchar('\n');
-    fflush(stdout);
-  }
-#endif
+
 }
 
 static void
@@ -1105,31 +1151,22 @@ contblock_sweep_phase(void) {
   ncb = 0;
 
   for (v=contblock_list_head;v;v=v->next) {
+    bool z;
 
     s=CB_DATA_START(v);
     e=(void *)v+v->in_use*PAGESIZE;
 
-    for (p = s;  p < e;) {
-      if (get_mark_bit(v,p)) {
-	/* SGC cont pages: cont blocks must be no smaller than
-	   sizeof(struct contblock), and must not have a sweep
-	   granularity greater than this amount (e.g. CPTR_SIZE) if
-	   contblock leaks are to be avoided.  Used to be aligned at
-	   PTR_ALIGN. CM 20030827 */
-	p += CPTR_SIZE;
-	continue;
-      }
-      q = p + CPTR_SIZE;
-      while (q < e) {
-	if (!get_mark_bit(v,q)) {
-	  q += CPTR_SIZE;
-	  continue;
-	}
-	break;
-      }
-      insert_contblock(p, q - p);
-      p = q + CPTR_SIZE;
+    z=get_mark_bit(v,s);
+    for (p=s;p<e;) {
+      q=get_bits(CB_MARK_START(v),v,p);
+      if (!z)
+	insert_contblock(p,q-p);
+      z=1-z;
+      p=q;
     }
+
+    bzero(CB_MARK_START(v),CB_SGCF_START(v)-CB_MARK_START(v));
+
   }
 #ifdef DEBUG
   if (debug) {
@@ -1138,6 +1175,9 @@ contblock_sweep_phase(void) {
     fflush(stdout);
   }
 #endif
+  
+  sweep_link_array();
+
 }
 
 
@@ -1170,7 +1210,6 @@ void
 GBC(enum type t) {
 
   long i,j;
-  struct apage *pp, *qq;
 #ifdef SGC
   int in_sgc = sgc_enabled;
 #endif
@@ -1178,6 +1217,13 @@ GBC(enum type t) {
   int tm=0;
 #endif
   
+  BEGIN_NO_INTERRUPT;
+
+  if (t==t_other) {
+    collect_both=1;
+    t=t_contiguous;
+  }
+
   if (in_signal_handler && t == t_relocatable)
     error("cant gc relocatable in signal handler");
   
@@ -1189,7 +1235,7 @@ GBC(enum type t) {
   interrupt_enable = FALSE;
   
   if (saving_system)
-    {t = t_contiguous; gc_time = -1;
+    {t = t_relocatable; gc_time = -1;
 #ifdef SGC
     if(sgc_enabled) sgc_quit();
 #endif    
@@ -1232,9 +1278,9 @@ GBC(enum type t) {
   
   maxpage = page(heap_end);
   
-  if ((int)t >= (int)t_contiguous) {
+  if (COLLECT_RELBLOCK_P) {
 
-    i=rb_pointer-REAL_RB_START;/*FIXME*/
+    i=rb_pointer-REAL_RB_START+PAGESIZE;/*FIXME*/
 
 #ifdef SGC
     if (sgc_enabled==0)
@@ -1263,14 +1309,6 @@ GBC(enum type t) {
     i = (rb_end < (rb_start1 + i) ? (rb_start1 + i) : rb_end) - heap_end;
     alloc_page(-(i + PAGESIZE - 1)/PAGESIZE);
     
-    { 
-      struct pageinfo *v;
-      for (v=contblock_list_head;v;v=v->next) {
-	void *p=pagetochar(page(v)),*pe=p+mbytes(v->in_use);
-	bzero(p,pe-p);
-      }
-    }
-
   }
   
 #ifdef DEBUG
@@ -1320,6 +1358,30 @@ GBC(enum type t) {
   }
 #endif
   
+  if (COLLECT_RELBLOCK_P) {
+    
+    if (rb_start < rb_start1) {
+      j = (rb_pointer-rb_start + PAGESIZE - 1)/PAGESIZE;
+      memcpy(rb_start,rb_start1,j*PAGESIZE);
+    }
+    
+#ifdef SGC
+    if (sgc_enabled)
+      wrimap=(void *)sSAwritableA->s.s_dbind->v.v_self;
+#endif
+
+#ifdef SGC
+    /* we don't know which pages have relblock on them */
+    if(sgc_enabled) {
+      fixnum i;
+      for (i=page(rb_start);i<page(rb_pointer+PAGESIZE-1);i++)
+	massert(IS_WRITABLE(i));
+    }    
+#endif		
+    rb_limit = rb_end - 2*RB_GETA;
+    
+  }
+
   if (t == t_contiguous) {
 #ifdef DEBUG
     if (debug) {
@@ -1342,35 +1404,6 @@ GBC(enum type t) {
 #endif
   }
   
-  if ((int)t >= (int)t_contiguous) {
-    
-    if (rb_start < rb_start1) {
-      j = (rb_pointer-rb_start + PAGESIZE - 1)/PAGESIZE;
-      pp = (struct apage *)rb_start;
-      qq = (struct apage *)rb_start1;
-      for (i = 0;  i < j;  i++)
-	*pp++ = *qq++;
-    }
-    
-#ifdef SGC
-      if (sgc_enabled)
-	wrimap=(void *)sSAwritableA->s.s_dbind->v.v_self;
-#endif
-
-#ifdef SGC
-    /* we don't know which pages have relblock on them */
-    if(sgc_enabled) {
-      fixnum i;
-      for (i=page(rb_start);i<page(rb_pointer+PAGESIZE-1);i++)
-	massert(IS_WRITABLE(i));
-    }    
-#endif		
-    rb_limit = rb_end - 2*RB_GETA;
-    
-    sweep_link_array();
-
-  }
-  
 #ifdef DEBUG
   if (debug) {
     for (i = 0, j = 0;  i < (int)t_end;  i++) {
@@ -1429,8 +1462,10 @@ GBC(enum type t) {
     
   }
 
- 
-  
+  collect_both=0;
+
+  END_NO_INTERRUPT;
+
   CHECK_INTERRUPT;
 
   /* ttss(); */
@@ -1457,8 +1492,13 @@ FFN(siLheap_report)(void) {
   vs_push(make_fixnum(CSTACK_ALIGNMENT));
   vs_push(make_fixnum(abs(cs_limit-cs_org)));/*CSSIZE*/
 #if defined(IM_FIX_BASE) && defined(IM_FIX_LIM)
+#ifdef LOW_IM_FIX
+  vs_push(make_fixnum(-LOW_IM_FIX));
+  vs_push(make_fixnum(1UL<<LOW_SHFT));
+#else
   vs_push(make_fixnum(IM_FIX_BASE));
   vs_push(make_fixnum(IM_FIX_LIM));
+#endif
 #else  
   vs_push(make_fixnum(0));
   vs_push(make_fixnum(0));
@@ -1473,10 +1513,6 @@ FFN(siLroom_report)(void) {
   
   check_arg(0);
   
-  /*
-    GBC(t_contiguous);
-  */
-  
   vs_check_push(make_fixnum(real_maxpage-first_data_page));
   vs_push(make_fixnum(available_pages));
   vs_push(make_fixnum(ncbpage));
@@ -1487,6 +1523,7 @@ FFN(siLroom_report)(void) {
   vs_push(make_fixnum(rb_pointer - rb_start));
   vs_push(make_fixnum(rb_end - rb_pointer));
   vs_push(make_fixnum(nrbpage));
+  vs_push(make_fixnum(maxrbpage));
   vs_push(make_fixnum(rbgbccount));
   for (i = 0;  i < (int)t_end;  i++) {
     if (tm_table[i].tm_type == (enum type)i) {
@@ -1530,8 +1567,9 @@ copy_relblock(char *p, int s)
  rb_pointer += s;
  rb_pointer1 += s;
  
- while (--s >= 0)
-   { *q++ = *p++;}
+ memcpy(q,p,s);
+ /* while (--s >= 0) */
+ /*   { *q++ = *p++;} */
  
  return res;
 }
@@ -1544,8 +1582,9 @@ mark_contblock(void *p, int s) {
   STATIC char *x, *y;
   struct pageinfo *v;
   
-  if (!MAYBE_DATA_P(p))
+  if (NULL_OR_ON_C_STACK(p))
     return;
+
   q = p + s;
   /* SGC cont pages: contblock pages must be no smaller than
      sizeof(struct contblock).  CM 20030827 */
@@ -1555,22 +1594,23 @@ mark_contblock(void *p, int s) {
 #ifdef SGC
   if (!sgc_enabled || (v->sgc_flags&SGC_PAGE_FLAG))
 #endif
-    for (;  x < y;  x+=CPTR_SIZE)
-      set_mark_bit(v,x);
+    set_mark_bits(v,x,y);
 }
 
-DEFUN_NEW("GBC",object,fLgbc,LISP
-       ,1,1,NONE,OO,OO,OO,OO,(object x0),"")
+DEFUN_NEW("GBC",object,fLgbc,LISP,1,1,NONE,OO,OO,OO,OO,(object x0),"") {
 
-{
   /* 1 args */
   
   if (x0 == Ct)
-    GBC(t_contiguous);
+    GBC(t_other);
   else if (x0 == Cnil)
     GBC(t_cons);
-  else
-    { x0 = small_fixnum(1);	GBC(t_relocatable);}
+  else if (eql(small_fixnum(0),x0))
+    GBC(t_contiguous);
+  else {
+    x0 = small_fixnum(1);
+    GBC(t_relocatable);
+  }
   RETURN1(x0);
 }
 
