@@ -149,7 +149,7 @@ mbrk(void *v) {
 int
 update_real_maxpage(void) {
 
-  ufixnum i,j;
+  ufixnum i,j,k;
   void *end,*cur;
 #ifdef __MINGW32__
   static fixnum n;
@@ -177,11 +177,23 @@ update_real_maxpage(void) {
 #endif
 
   available_pages=real_maxpage-first_data_page;
-  for (i=t_start;i<t_other;i++)
-    available_pages-=tm_table[i].tm_type==t_relocatable ? 2*tm_table[i].tm_maxpage : tm_table[i].tm_maxpage;
+  for (i=t_start,j=0;i<t_other;i++) {
+    k=tm_table[i].tm_maxpage;
+    if (tm_table[i].tm_type==t_relocatable)
+      k*=2;
+    else
+      j+=k;
+    available_pages-=k;
+  }
   resv_pages=40<available_pages ? 40 : available_pages;
   available_pages-=resv_pages;
-    
+
+  new_holepage=available_pages/starting_hole_div;
+  k=available_pages/20;
+  j*=starting_relb_heap_mult;
+  j=j<k ? j : k;
+  if (maxrbpage<j)
+    set_tm_maxpage(tm_table+t_relocatable,j);
 
   return 0;
 
@@ -201,7 +213,7 @@ minimize_image(void) {
   
   if (in_sgc) sgc_quit();
   holepage=new_holepage=1;
-  GBC(t_contiguous);
+  GBC(t_relocatable);
   if (in_sgc) sgc_start();
   new = (void *)(((((ufixnum)rb_pointer)+ PAGESIZE-1)/PAGESIZE)*PAGESIZE);
   core_end = new;
@@ -214,7 +226,7 @@ minimize_image(void) {
 #endif
   
 #if defined(BSD) || defined(ATT)  
-  brk(core_end);
+  mbrk(core_end);
 #endif
   
   cbgbccount = tm_table[t_contiguous].tm_adjgbccnt = tm_table[t_contiguous].tm_opt_maxpage = 0;
@@ -233,7 +245,7 @@ DEFUN_NEW("SET-LOG-MAXPAGE-BOUND",fixnum,fSset_log_maxpage_bound,SI,1,1,NONE,II,
 
   l=l<def ? l : def;
   end=data_start+(1L<<l)-PAGESIZE;
-  GBC(t_contiguous);
+  GBC(t_relocatable);
   dend=heap_end+PAGESIZE+(((rb_pointer-REAL_RB_START)+PAGESIZE-1)&(-PAGESIZE));
   if (end >= dend) {
     minimize_image();
@@ -465,7 +477,13 @@ main(int argc, char **argv, char **envp) {
 #ifdef _WIN32
 	  detect_wine();
 #endif
-	  
+
+#ifdef LD_BIND_NOW /*FIXME currently mips only, verify that these two
+		     requirements are the same*/
+	  reinit_gmp();
+#endif
+
+
 		if (saving_system) {
 
 			saving_system = FALSE;
@@ -475,19 +493,6 @@ main(int argc, char **argv, char **envp) {
 #ifdef INIT_CORE_END
 			INIT_CORE_END
 #endif			  
-			  /* if (available_pages/20>nrbpage) */
-			  /*   add_pages(tm_table+t_relocatable,(available_pages/20-nrbpage)<<PAGEWIDTH); */
-			  /* { */
-			  /*   int in_sgc=sgc_enabled; */
-			  /*   if (in_sgc) sgc_quit(); */
-			  /*   /\* holepage=available_pages/10; *\/ */
-			  /*   /\* if (available_pages/20>nrbpage) nrbpage=available_pages/20; *\/ */
-			  /*   GBC_enable = TRUE; */
-			  /*   GBC(t_relocatable); */
-			  /*   if (in_sgc) sgc_start(); */
-			  /* } */
-			  /* alloc_page(-((available_pages/5) + nrbpage)); */
-			  alloc_page(-(holepage + nrbpage));
 		}
 
 		initflag = FALSE;
@@ -627,6 +632,8 @@ initlisp(void) {
 	
 	gcl_init_alloc();
 	
+	Cnil->fw=0;
+	set_type_of(Cnil,t_symbol);
  	Cnil->c.c_cdr=Cnil;
 	Cnil_body.s.s_dbind = Cnil;
 	Cnil_body.s.s_sfdef = NOT_SPECIAL;
@@ -638,6 +645,7 @@ initlisp(void) {
 	Cnil_body.s.s_stype = (short)stp_constant;
 	Cnil_body.s.s_mflag = FALSE;
 	
+	Ct->fw=0;
 	set_type_of(Ct,t_symbol);
 	Ct_body.s.s_dbind = Ct;
 	Ct_body.s.s_sfdef = NOT_SPECIAL;
@@ -1098,16 +1106,8 @@ FFN(siLsave_system)(void) {
 #endif	
     
   saving_system = TRUE;
-  {
 
-    fixnum old_nrbpage=nrbpage;
-
-    minimize_image();
-
-    rb_limit=rb_end=REAL_RB_START+old_nrbpage*PAGESIZE;
-    set_tm_maxpage(tm_table+t_relocatable,nrbpage=old_nrbpage);
-
-  }
+  minimize_image();
 
   Lsave();
   saving_system = FALSE;
