@@ -110,20 +110,18 @@ joe() {;}
 /*   sgc_mark_object1(x); */
 /* } */
 
-static void
+inline void
 sgc_mark_cons(object x) {
   
-  object d;
-
   do {
-    d=x->c.c_cdr;
+    object d=x->c.c_cdr;
     mark(x);
     sgc_mark_object(x->c.c_car);
     x=d;
-    if (!IS_WRITABLE(page(x)) || is_marked_or_free(x))
+    if (!IS_WRITABLE(page(x)) || is_marked_or_free(x))/*catches Cnil*/
       return;
-  } while (type_of(x)==t_cons);
-  sgc_mark_object1(x);
+  } while (cdr_listp(x));
+  sgc_mark_object(x);
 
 }
 
@@ -162,13 +160,17 @@ sgc_mark_object1(object x) {
   if(x==sdebug) joe1();
 #endif
   
-  if ((tp=type_of(x))==t_cons) {
+  tp=type_of(x);
+
+  if (tp==t_cons) {
     sgc_mark_cons(x);
     return;
   }
 
   mark(x);
+
   switch (tp) {
+
   case t_fixnum:
     break;
     
@@ -200,15 +202,11 @@ sgc_mark_object1(object x) {
     if (x->s.s_self == NULL)
       break;
     /* to do */
-    if (what_to_collect >= t_contiguous) {
-      if (inheap(x->s.s_self)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(x->s.s_self,
-			 x->s.s_fillp);
-      } else if(SGC_RELBLOCK_P(x->s.s_self))
-	x->s.s_self =
-	  copy_relblock(x->s.s_self, x->s.s_fillp);
-    }
+    if (inheap(x->s.s_self)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(x->s.s_self,x->s.s_fillp);
+    } else if (SGC_RELBLOCK_P(x->s.s_self) && COLLECT_RELBLOCK_P)
+      x->s.s_self = copy_relblock(x->s.s_self, x->s.s_fillp);
     break;
     
   case t_package:
@@ -217,23 +215,14 @@ sgc_mark_object1(object x) {
     sgc_mark_object(x->p.p_shadowings);
     sgc_mark_object(x->p.p_uselist);
     sgc_mark_object(x->p.p_usedbylist);
-    if (what_to_collect != t_contiguous)
-      break;
-    if (x->p.p_internal != NULL)
-      mark_contblock((char *)(x->p.p_internal),
-		     x->p.p_internal_size*sizeof(object));
-    if (x->p.p_external != NULL)
-      mark_contblock((char *)(x->p.p_external),
-		     x->p.p_external_size*sizeof(object));
-    break;
-    
-  case t_cons:
-    /*
-      sgc_mark_object(x->c.c_car);
-      x = x->c.c_cdr;
-      goto BEGIN;
-    */
-    sgc_mark_cons(x);
+    if (what_to_collect == t_contiguous) {
+      if (x->p.p_internal != NULL)
+	mark_contblock((char *)(x->p.p_internal),
+		       x->p.p_internal_size*sizeof(object));
+      if (x->p.p_external != NULL)
+	mark_contblock((char *)(x->p.p_external),
+		       x->p.p_external_size*sizeof(object));
+    }
     break;
     
   case t_hashtable:
@@ -247,29 +236,22 @@ sgc_mark_object1(object x) {
 	sgc_mark_object(x->ht.ht_self[i].hte_value);
       }
     }
-    if ((short)what_to_collect >= (short)t_contiguous) {
-      if (inheap(x->ht.ht_self)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)(x->ht.ht_self),
-			 j * sizeof(struct htent));
-      } else if(SGC_RELBLOCK_P(x->ht.ht_self))
-	x->ht.ht_self=(void *)copy_relblock((char *)x->ht.ht_self,j*sizeof(struct htent));;
-    }
+    if (inheap(x->ht.ht_self)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock((char *)(x->ht.ht_self),j * sizeof(struct htent));
+    } else if (SGC_RELBLOCK_P(x->ht.ht_self) && COLLECT_RELBLOCK_P)
+      x->ht.ht_self=(void *)copy_relblock((char *)x->ht.ht_self,j*sizeof(struct htent));;
     break;
     
   case t_array:
     if ((x->a.a_displaced) != Cnil)
       sgc_mark_displaced_field(x);
-    if (what_to_collect >= t_contiguous &&
-	x->a.a_dims != NULL) {
+    if (x->a.a_dims != NULL) {
       if (inheap(x->a.a_dims)) {
 	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)(x->a.a_dims),
-			 sizeof(int)*x->a.a_rank);
-      } else  if(SGC_RELBLOCK_P(x->a.a_dims))
-	x->a.a_dims = (int *)
-	  copy_relblock((char *)(x->a.a_dims),
-			sizeof(int)*x->a.a_rank);
+	  mark_contblock((char *)(x->a.a_dims),sizeof(int)*x->a.a_rank);
+      } else if (SGC_RELBLOCK_P(x->a.a_dims) && COLLECT_RELBLOCK_P)
+	x->a.a_dims = (int *) copy_relblock((char *)(x->a.a_dims),sizeof(int)*x->a.a_rank);
     }
     if ((enum aelttype)x->a.a_elttype == aet_ch)
       goto CASE_STRING;
@@ -287,9 +269,8 @@ sgc_mark_object1(object x) {
     switch((enum aelttype)x->a.a_elttype){
     case aet_lf:
       j= sizeof(longfloat)*x->lfa.lfa_dim;
-      if ((what_to_collect >= t_contiguous) &&
-	  !(inheap(cp)) && SGC_RELBLOCK_P(x->a.a_self))
-	ROUND_RB_POINTERS_DOUBLE;
+      if ((COLLECT_RELBLOCK_P) && !(inheap(cp)) && SGC_RELBLOCK_P(x->a.a_self))
+	ROUND_RB_POINTERS_DOUBLE;/*FIXME gc space violation*/
       break;
     case aet_char:
     case aet_uchar:
@@ -321,21 +302,17 @@ sgc_mark_object1(object x) {
     cp = (char *)p;
     j *= sizeof(object);
   COPY:
-    if (what_to_collect >= t_contiguous) {
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      } else if (!SGC_RELBLOCK_P(cp)) 
-	;
-      else if (x->a.a_displaced == Cnil) {
+    if (inheap(cp)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(cp, j);
+    } else if (SGC_RELBLOCK_P(cp) && COLLECT_RELBLOCK_P) {
+      if (x->a.a_displaced == Cnil) {
 #ifdef HAVE_ALLOCA
 	if (!NULL_OR_ON_C_STACK(cp))  /* only if body of array not on C stack */
 #endif			  
 	  x->a.a_self = (object *)copy_relblock(cp, j);
-      }
-      else if (x->a.a_displaced->c.c_car == Cnil) {
-	i = (long)(object *)copy_relblock(cp, j)
-	  - (long)(x->a.a_self);
+      } else if (x->a.a_displaced->c.c_car == Cnil) {
+	i = (long)(object *)copy_relblock(cp, j) - (long)(x->a.a_self);
 	adjust_displaced(x, i);
       }
     }
@@ -354,52 +331,17 @@ sgc_mark_object1(object x) {
     if (TYPE_MAP(page(x->big.big_self)) < t_contiguous)
 	printf("bad body for %x (%x)\n",x,cp);
 #endif
-#ifndef GMP
-    if (what_to_collect >= t_contiguous) {
-      j = x->big.big_length;
-      cp = (char *)x->big.big_self;
-      if (cp == NULL)
-	break;
-      if  (j != lg(MP(x))  &&
-	   /* we don't bother to zero this register,
-	      and its contents may get over written */
-	   ! (x ==  big_register_1 &&
-	      (int)(cp) <= top &&
-	      (int) cp >= bot))
-	
-	printf("bad length 0x%x ",x);
-      j = j * sizeof(int);
-      
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      } else {
-	if (SGC_RELBLOCK_P(cp))
-	  x->big.big_self = (plong *)copy_relblock(cp, j);}}
-#endif /* no gmp */
 #ifndef GMP_USE_MALLOC
-    if (what_to_collect >= t_contiguous) {
-      j = MP_ALLOCATED(x);
-      cp = (char *)MP_SELF(x);
-      if (cp == 0)
-	break;
-#ifdef PARI
-      if (j != lg(MP(x))  &&
-	  /* we don't bother to zero this register,
-	     and its contents may get over written */
-	  ! (x == big_register_1 &&
-	     (int)(cp) <= top &&
-	     (int) cp >= bot))
-	printf("bad length 0x%x ",x);
-#endif
-      j = j * MP_LIMB_SIZE;
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      } else 
-	if (SGC_RELBLOCK_P(cp))
-	  MP_SELF(x) = (void *) copy_relblock(cp, j);
-    }
+    j = MP_ALLOCATED(x);
+    cp = (char *)MP_SELF(x);
+    if (cp == 0)
+      break;
+    j = j * MP_LIMB_SIZE;
+    if (inheap(cp)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(cp, j);
+    } else if (SGC_RELBLOCK_P(cp) && COLLECT_RELBLOCK_P)
+      MP_SELF(x) = (void *) copy_relblock(cp, j);
 #endif /* not GMP_USE_MALLOC */
     break;
     
@@ -414,13 +356,11 @@ sgc_mark_object1(object x) {
       break;
     
   COPY_STRING:
-    if (what_to_collect >= t_contiguous) {
-      if (inheap(cp)) {
-	if (what_to_collect == t_contiguous)
-	  mark_contblock(cp, j);
-      }
-      else if (!SGC_RELBLOCK_P(cp)) ;
-      else if (x->st.st_displaced == Cnil)
+    if (inheap(cp)) {
+      if (what_to_collect == t_contiguous)
+	mark_contblock(cp, j);
+    } else if (SGC_RELBLOCK_P(cp) && COLLECT_RELBLOCK_P) {
+      if (x->st.st_displaced == Cnil)
 	x->st.st_self = copy_relblock(cp, j);
       else if (x->st.st_displaced->c.c_car == Cnil) {
 	i = copy_relblock(cp, j) - cp;
@@ -456,16 +396,11 @@ sgc_mark_object1(object x) {
       for (i = 0, j = S_DATA(def)->length;  i < j;  i++)
 	if (s_type[i]==0 && ON_WRITABLE_PAGE(&STREF(object,x,s_pos[i])))
 	  sgc_mark_object(STREF(object,x,s_pos[i]));
-      if (what_to_collect >= t_contiguous) {
-	if (inheap(x->str.str_self)) {
-	  if (what_to_collect == t_contiguous)
-	    mark_contblock((char *)p,
-			   S_DATA(def)->size);
-	  
-	} else if(SGC_RELBLOCK_P(p))
-	  x->str.str_self = (object *)
-	    copy_relblock((char *)p, S_DATA(def)->size);
-      }
+      if (inheap(x->str.str_self)) {
+	if (what_to_collect == t_contiguous)
+	  mark_contblock((char *)p,S_DATA(def)->size);
+      } else if (SGC_RELBLOCK_P(p) && (COLLECT_RELBLOCK_P))
+	x->str.str_self = (object *) copy_relblock((char *)p, S_DATA(def)->size);
     }
     break;
     
@@ -522,21 +457,33 @@ sgc_mark_object1(object x) {
     }
     break;
     
+#define SGC_MARK_CP(a_,b_) {fixnum _t=(b_);if (inheap((a_))) {\
+	if (what_to_collect == t_contiguous) mark_contblock((void *)(a_),_t); \
+      } else if (SGC_RELBLOCK_P((a_)) && COLLECT_RELBLOCK_P) (a_)=(void *)copy_relblock((void *)(a_),_t);}
+
+#define SGC_MARK_MP(a_) {if ((a_)->_mp_d) SGC_MARK_CP((a_)->_mp_d,(a_)->_mp_alloc*MP_LIMB_SIZE);}
+
   case t_random:
+    SGC_MARK_MP(x->rnd.rnd_state._mp_seed);
+#if __GNU_MP_VERSION < 4 || (__GNU_MP_VERSION == 4 && __GNU_MP_VERSION_MINOR < 2)
+    if (x->rnd.rnd_state._mp_algdata._mp_lc) {
+      SGC_MARK_MP(x->rnd.rnd_state._mp_algdata._mp_lc->_mp_a);
+      if (!x->rnd.rnd_state._mp_algdata._mp_lc->_mp_m2exp) SGC_MARK_MP(x->rnd.rnd_state._mp_algdata._mp_lc->_mp_m);
+      SGC_MARK_CP(x->rnd.rnd_state._mp_algdata._mp_lc,sizeof(*x->rnd.rnd_state._mp_algdata._mp_lc));
+    }
+#endif
     break;
     
   case t_readtable:
     if (x->rt.rt_self == NULL)
       break;
     if (what_to_collect == t_contiguous)
-      mark_contblock((char *)(x->rt.rt_self),
-		     RTABSIZE*sizeof(struct rtent));
+      mark_contblock((char *)(x->rt.rt_self),RTABSIZE*sizeof(struct rtent));
     for (i = 0;  i < RTABSIZE;  i++) {
       sgc_mark_object(x->rt.rt_self[i].rte_macro);
       if (x->rt.rt_self[i].rte_dtab != NULL) {
 	if (what_to_collect == t_contiguous)
-	  mark_contblock((char *)(x->rt.rt_self[i].rte_dtab),
-			 RTABSIZE*sizeof(object));
+	  mark_contblock((char *)(x->rt.rt_self[i].rte_dtab),RTABSIZE*sizeof(object));
 	for (j = 0;  j < RTABSIZE;  j++)
 	  sgc_mark_object(x->rt.rt_self[i].rte_dtab[j]);
       }
@@ -577,12 +524,7 @@ sgc_mark_object1(object x) {
       while(i-- > 0)
 	sgc_mark_object(x->cfd.cfd_self[i]);
     }
-    if (x->cfd.cfd_start == NULL)
-      break;
     if (what_to_collect == t_contiguous) {
-      if (!MAYBE_DATA_P((x->cfd.cfd_start)) ||
-	  get_mark_bit(get_pageinfo(x->cfd.cfd_start),x->cfd.cfd_start))
-	break;
       mark_contblock(x->cfd.cfd_start, x->cfd.cfd_size);
       mark_link_array(x->cfd.cfd_start,x->cfd.cfd_start+x->cfd.cfd_size);
     }
@@ -594,8 +536,7 @@ sgc_mark_object1(object x) {
     if (x->cc.cc_turbo!=NULL) sgc_mark_object(*(x->cc.cc_turbo-1));
     if (what_to_collect == t_contiguous) {
       if (x->cc.cc_turbo != NULL)
-	mark_contblock((char *)(x->cc.cc_turbo-1),
-		       (1+fix(*(x->cc.cc_turbo-1)))*sizeof(object));
+	mark_contblock((char *)(x->cc.cc_turbo-1),(1+fix(*(x->cc.cc_turbo-1)))*sizeof(object));
     }
     break;
     
@@ -646,22 +587,24 @@ sgc_mark_phase(void) {
     }
   }
   
+  /* mark all non recent data on writable contiguous pages */
+  if (what_to_collect == t_contiguous)
+    for (v=contblock_list_head;v;v=v->next)
+      if (v->sgc_flags&SGC_PAGE_FLAG) {
+	void *s=CB_DATA_START(v),*e=CB_DATA_END(v),*p,*q;
+	bool z=get_sgc_bit(v,s);
+	for (p=s;p<e;) {
+	  q=get_sgc_bits(v,p);
+	  if (!z)
+	    set_mark_bits(v,p,q);
+	  z=1-z;
+	  p=q;
+	}
+      }
+	    
   mark_stack_carefully(vs_top-1,vs_org,0);
-  /* clear_stack(vs_top,vs_limit); */
   mark_stack_carefully(MVloc+(sizeof(MVloc)/sizeof(object)),MVloc,0);
-  /* 
-     for (p = vs_org;  p < vs_top;  p++) {
-     if (p && (inheap(*p)))
-     sgc_mark_object(*p);
-     }
-  */
-#ifdef DEBUG
-  if (debug) {
-    printf("value stack marked\n");
-    fflush(stdout);
-  }
-#endif
-  
+
   for (bdp = bds_org;  bdp<=bds_top;  bdp++) {
     sgc_mark_object(bdp->bds_sym);
     sgc_mark_object(bdp->bds_val);
@@ -713,7 +656,7 @@ sgc_mark_phase(void) {
 
 static void
 sgc_sweep_phase(void) {
-  STATIC long i, j, k;
+  STATIC long j, k;
   STATIC object x;
   STATIC char *p;
   STATIC struct typemanager *tm;
@@ -721,46 +664,22 @@ sgc_sweep_phase(void) {
   int size;
   STATIC struct pageinfo *v;
   
-#ifdef DEBUG
-  if (debug)
-    printf("type map\n");
-#endif
   for (v=cell_list_head;v;v=v->next) {
-    enum type tp=v->type;
 
-    i=page(v);
+    tm = tm_of((enum type)v->type);
+    
+    if (!WRITABLE_PAGE_P(page(v))) 
+      continue;
 
-    if (tp == t_contiguous) {
-      if (debug) {
-	printf("-");
-	continue;
-      }
-    }
-    if (tp >= t_end)
-      continue;
-    
-    tm = tm_of((enum type)tp);
-    
-    /*
-      general sweeper
-    */
-    
-#ifdef DEBUG
-    if (debug) {
-      printf("%c", tm->tm_name[0]);
-      /*
-	fflush(stdout);
-      */
-    }
-#endif
-    if (!WRITABLE_PAGE_P(i)) 
-      continue;
-    p = pagetochar(i);
+    p = pagetochar(page(v));
     f = tm->tm_free;
     k = 0;
     size=tm->tm_size;
-    if (v->sgc_flags&SGC_PAGE_FLAG) {/*(SGC_PAGE_P(i))*/
+
+    if (v->sgc_flags&SGC_PAGE_FLAG) {
+
       for (j = tm->tm_nppage; --j >= 0;  p += size) {
+
 	x = (object)p;
 	
 	if (is_free(x))
@@ -769,113 +688,66 @@ sgc_sweep_phase(void) {
 	  unmark(x);
 	  continue;
 	}
-	if (pageinfo(x)->type!=t_cons && x->d.s == SGC_NORMAL)
+
+	if (TYPEWORD_TYPE_P(pageinfo(x)->type) && x->d.s == SGC_NORMAL)
 	  continue;
 	
 	/* it is ok to free x */
 	
-#ifdef OLD_DISPLACE
-	/* old_displace: from might be free, to not */
-	if(x->d.t >=t_array && x->d.t <= t_bitvector) {
-	  /*			case t_array:
-				case t_vector:
-				case t_string:
-				case t_bitvector:
-	  */			
-	  if (x->a.a_displaced->c.c_car != Cnil) {
-	    undisplace(x);
-	    /* The cons x->a.a_displaced cons has been saved,
-	       so as to save the pointer to x->a.a_displaced->c.c_car;
-	       However any arrays in its cdr, must have been
-	       freed, or we would not be freeing x.   To avoid
-	       having a cons with trash in the cdr we set the
-	       cdr to nil
-	    */				    
-	    x->a.a_displaced->c.c_cdr = Cnil;
-	  }
-	}
-#endif /* OLD_DISPLACE */
-#ifdef GMP_USE_MALLOC			
-	if (x->d.t == t_bignum) 
-	  mpz_clear(MP(x));
-#endif
-	
 	SET_LINK(x,f);
 	make_free(x);
-	if (pagetoinfo(i)->type!=t_cons) x->d.s = SGC_RECENT;
+	if (TYPEWORD_TYPE_P(v->type)) x->d.s = SGC_RECENT;
 	f = x;
 	k++;
+
       }
       tm->tm_free = f;
       tm->tm_nfree += k;
-      pagetoinfo(i)->in_use-=k;
-    }
-    else /*non sgc_page */
+      v->in_use-=k;
+
+    } else /*non sgc_page */
       for (j = tm->tm_nppage; --j >= 0;  p += size) {
 	x = (object)p;
-	if (is_marked(x) && !is_free(x)) {unmark(x);}
+	if (is_marked(x) && !is_free(x)) {
+	  unmark(x);
+	}
       }
     
   }
-#ifdef DEBUG
-  if (debug) {
-    putchar('\n');
-    fflush(stdout);
-  }
-#endif
 }
 
 
 static void
 sgc_contblock_sweep_phase(void) {
 
-  STATIC long i, j;
   STATIC char *s, *e, *p, *q;
-  STATIC struct contblock *cbp;
   STATIC struct pageinfo *v;
   
   cb_pointer = NULL;
   ncb = 0;
   for (v=contblock_list_head;v;v=v->next) {
+    bool z;
 
     if (!(v->sgc_flags&SGC_PAGE_FLAG)) continue;
 
-    i=page(v);
-    j=i+v->in_use;
-    
-    s=pagetochar(i)+mbytes(v->in_use);
-    e=(void *)pagetoinfo(j);
+    s=CB_DATA_START(v);
+    e=CB_DATA_END(v);
 
-    for (p = s;  p < e;) {
-      if (get_mark_bit(v,p)) {
-	/* SGC cont pages: cont blocks must be no smaller than
-	   sizeof(struct contblock), and must not have a sweep
-	   granularity greater than this amount (e.g. CPTR_SIZE) if
-	   contblock leaks are to be avoided.  Used to be aligned at
-	   PTR_ALIGN. CM 20030827 */
-	p += CPTR_SIZE;
-	continue;
-      }
-      q = p + CPTR_SIZE;
-      while (q < e) {
-	if (!get_mark_bit(v,q)) {
-	  q += CPTR_SIZE;
-	  continue;
-	}
-	break;
-      }
-      insert_contblock(p, q - p);
-      p = q + CPTR_SIZE;
+    z=get_mark_bit(v,s);
+    for (p=s;p<e;) {
+      q=get_mark_bits(v,p);
+      if (!z)
+	insert_contblock(p,q-p);
+      z=1-z;
+      p=q;
     }
-    /* i = j + 1; */
+
+    bzero(CB_MARK_START(v),CB_SGCF_START(v)-CB_MARK_START(v));
+
   }
-#ifdef DEBUG
-  if (debug) {
-    for (cbp = cb_pointer; cbp != NULL; cbp = cbp->cb_link)
-      printf("%d-byte contblock\n", cbp->cb_size);
-    fflush(stdout);
-  }
-#endif
+
+  sweep_link_array();
+
 }
 
 
@@ -1169,7 +1041,8 @@ memprotect_test_reset(void) {
 #define WSGC(tm) ({struct typemanager *_tm=tm;long _t=MMAX(MMIN(_tm->tm_opt_maxpage,_tm->tm_npage),_tm->tm_sgc);_t*scale;})
 /* If opt_maxpage is set, add full pages to the sgc set if needed
    too. 20040804 CM*/
-#define FSGC(tm) (tm->tm_type==t_cons ? tm->tm_nppage : (tm->tm_opt_maxpage ? 0 : tm->tm_sgc_minfree))
+/* #define FSGC(tm) (tm->tm_type==t_cons ? tm->tm_nppage : (tm->tm_opt_maxpage ? 0 : tm->tm_sgc_minfree)) */
+#define FSGC(tm) (!TYPEWORD_TYPE_P(tm->tm_type) ? tm->tm_nppage : tm->tm_sgc_minfree)
 
 DEFVAR("*WRITABLE*",sSAwritableA,SI,Cnil,"");
 
@@ -1249,9 +1122,12 @@ sgc_start(void) {
       long again=0,nfree = tm->tm_nfree;
       char *p=alloc_page(n);
       if (tm->tm_nfree > nfree) again=1;  /* gc freed some objects */
+      if (tm->tm_npage+n>tm->tm_maxpage)
+	if (!set_tm_maxpage(tm,tm->tm_npage+n))
+	  n=0;
       while (n-- > 0) {
 	/* (sgc_enabled=1,add_page_to_freelist(p,tm),sgc_enabled=0); */
-	(add_page_to_freelist(p,tm));
+	add_page_to_freelist(p,tm);
 	p += PAGESIZE;
       }
       if (again) 
@@ -1291,17 +1167,16 @@ sgc_start(void) {
     
     tm=tm_of(t_contiguous);
 
-    for (pi=contblock_list_head;pi;pi=pi->next) {
+    for (pi=contblock_list_head;pi && count<WSGC(tm);pi=pi->next) {
 
-      j=CB_DATA_SIZE(pi->in_use);
       p=CB_DATA_START(pi);
-      pe=p+j;
+      pe=CB_DATA_END(pi);
 
-      for (cbpp=&cb_pointer;*cbpp;cbpp=&(*cbpp)->cb_link)
+      for (cbpp=&cb_pointer,j=0;*cbpp;cbpp=&(*cbpp)->cb_link)
 	if ((void*)*cbpp>=p && (void *)*cbpp<pe)
-	  j-=(*cbpp)->cb_size;
+	  j+=(*cbpp)->cb_size;
 
-      if (j) continue;
+      if (j*tm->tm_nppage<FSGC(tm)*(CB_DATA_END(pi)-CB_DATA_START(pi))) continue;
 
       pi->sgc_flags=SGC_PAGE_FLAG;
       count+=pi->in_use;
@@ -1316,7 +1191,8 @@ sgc_start(void) {
       void *old_contblock_list_tail=contblock_list_tail;
 
       if (maxcbpage<ncbpage+z)
-	massert(set_tm_maxpage(tm_table+t_contiguous,ncbpage+z));
+	if (!set_tm_maxpage(tm_table+t_contiguous,ncbpage+z))
+	  z=0;
 
       add_pages(tm_table+t_contiguous,z);
 
@@ -1356,9 +1232,9 @@ sgc_start(void) {
     if (TM_BASE_TYPE_P(i) && (np=(tm=tm_of(i))->tm_sgc)) {
       object f=tm->tm_free ,x,y,next;
       fixnum count=0;
-      x=y=0;
+      x=y=OBJNULL;
       
-      while (f!=0) {
+      while (f!=OBJNULL) {
 	next=OBJ_LINK(f);
 #ifdef SDEBUG	     
 	if (!is_free(f))
@@ -1366,12 +1242,12 @@ sgc_start(void) {
 #endif
 	if (pageinfo(f)->sgc_flags&SGC_PAGE_FLAG) {
 	  SET_LINK(f,x);
-	  if (pageinfo(f)->type!=t_cons) f->d.s = SGC_RECENT;
+	  if (TYPEWORD_TYPE_P(pageinfo(f)->type)) f->d.s = SGC_RECENT;
 	  x=f;
 	  count++;
 	} else {
 	  SET_LINK(f,y);
- 	  if (pageinfo(f)->type!=t_cons) f->d.s = SGC_NORMAL;
+ 	  if (TYPEWORD_TYPE_P(pageinfo(f)->type)) f->d.s = SGC_NORMAL;
 	  y=f;
 	}
 	f=next;
@@ -1400,20 +1276,18 @@ sgc_start(void) {
 	pe=p+CB_DATA_SIZE(pi->in_use);
 	
 	for (cbpp=&cb_pointer;*cbpp;)
-	  if ((void *)*cbpp>=p && (void *)*cbpp<pe)
-	    *cbpp=(*cbpp)->cb_link;
-	  else
+	  if ((void *)*cbpp>=p && (void *)*cbpp<pe) {
+	    void *s=*cbpp,*e=s+(*cbpp)->cb_size,*l=(*cbpp)->cb_link;
+	    set_sgc_bits(pi,s,e);
+	    tmp_cb_pointer=cb_pointer;
+	    cb_pointer=new_cb_pointer;
+	    insert_contblock(s,e-s);
+	    new_cb_pointer=cb_pointer;
+	    cb_pointer=tmp_cb_pointer;
+	    *cbpp=l;
+	  } else
 	    cbpp=&(*cbpp)->cb_link;
-	
-	
-	tmp_cb_pointer=cb_pointer;
-	cb_pointer=new_cb_pointer;
-	/* SGC contblock pages: add whole pages to new list, p p-k, and
-	   p+i are guaranteed to be distinct when used. CM 20030827 */
-	insert_contblock(p,pe-p);
-	new_cb_pointer=cb_pointer;
-	cb_pointer=tmp_cb_pointer;
-	
+
       }
       
       /* SGC contblock pages: switch to new free SGC contblock list. CM
@@ -1453,6 +1327,14 @@ sgc_start(void) {
 	  SET_WRITABLE(page(pi)+i);
       else
 	tm_of(t_contiguous)->tm_alt_npage+=pi->in_use;
+#ifdef GCL_GPROF
+    {
+      extern object gprof_array;
+      if (gprof_array!=Cnil)
+	for (i=0;i<(gprof_array->st.st_fillp +PAGESIZE-1)/PAGESIZE;i++)
+	  SET_WRITABLE(page(gprof_array->st.st_self)+i);
+    }
+#endif
     for (i=page(heap_end);i<page(old_rb_start);i++)
 	SET_WRITABLE(i);
     tm_of(t_relocatable)->tm_alt_npage=page(rb_start)-page(old_rb_start);
@@ -1541,7 +1423,7 @@ sgc_quit(void) {
       object f,y;
       
       f=tm->tm_free;
-      if (f==0) 
+      if (f==OBJNULL) 
 	tm->tm_free=tm->tm_alt_free;
       else {
 	/* tack the alt_free onto the end of free */
@@ -1555,7 +1437,7 @@ sgc_quit(void) {
 	}
 	
 	count=0;
-	if (f=tm->tm_alt_free)
+	if (f==tm->tm_alt_free)
 	  while(y= F_LINK(f)) {
 	    if(y->d.s != SGC_NORMAL)
 	      printf("[alt_bad %d]",y);
@@ -1564,23 +1446,27 @@ sgc_quit(void) {
 	
 #endif
 	f=tm->tm_free;
-	while((y= (object) F_LINK(f)))
+	while((y= (object) F_LINK(f))!=OBJNULL)
 	  f=y;
 	F_LINK(f)= (long)(tm->tm_alt_free);
       }
       /* tm->tm_free has all of the free objects */
       tm->tm_nfree += tm->tm_alt_nfree;
       tm->tm_alt_nfree = 0;
-      tm->tm_alt_free = 0;
+      tm->tm_alt_free = OBJNULL;
       
     }
 
   /*FIXME*/
   /* remove the recent flag from any objects on sgc pages */
   for (v=cell_list_head;v;v=v->next) 
-    if (v->type==(tm=tm_of(v->type))->tm_type && v->type!=t_cons && v->sgc_flags & SGC_PAGE_FLAG)
+    if (v->type==(tm=tm_of(v->type))->tm_type && TYPEWORD_TYPE_P(v->type) && v->sgc_flags & SGC_PAGE_FLAG)
       for (p=pagetochar(page(v)),j=tm->tm_nppage;j>0;--j,p+=tm->tm_size)
 	((object) p)->d.s=SGC_NORMAL;
+
+  for (v=contblock_list_head;v;v=v->next) 
+    if (v->sgc_flags&SGC_PAGE_FLAG) 
+      bzero(CB_SGCF_START(v),CB_DATA_START(v)-CB_SGCF_START(v));
   
   {
     struct pageinfo *pi;
