@@ -242,7 +242,7 @@ truename(object pathname)
            FEerror ( "truename could not determine the current directory.", 1, "" ); 
         } 
 #else 
-        getcwd(current_directory,sizeof(current_directory)); 
+        massert(current_directory==getcwd(current_directory,sizeof(current_directory))); 
 #endif 
     
 	coerce_to_filename(pathname, filename);
@@ -346,7 +346,7 @@ truename(object pathname)
 		}
 	} else
 		sprintf(truefilename, "%s/%s", p, q);
-	chdir(current_directory);
+	massert(!chdir(current_directory));
 	vs_push(make_simple_string(truefilename));
 	pathname = coerce_to_pathname(vs_head);
 	vs_popp;
@@ -478,6 +478,13 @@ DEFUN_NEW("STAT",object,fSstat,SI,1,1,NONE,OO,OO,OO,OO,(object path),"") {
 
   bzero(filename,sizeof(filename));
   coerce_to_filename(path,filename);
+#ifdef __MINGW32__
+  {
+    char *p=filename+strlen(filename)-1;
+    for (;p>filename && *p=='/';p--)
+      *p=0;
+  }
+#endif
   if (lstat(filename,&ss))
     RETURN1(Cnil);
   else {/* ctime_r insufficiently portable */
@@ -774,58 +781,73 @@ LFD(Ldirectory)()
 }
 #endif
 
+#include <sys/types.h>
+#include <dirent.h>
 
-#ifdef DGUX
+DEFUN_NEW("OPENDIR",fixnum,fSopendir,SI,1,1,NONE,IO,OO,OO,OO,(object x),"") {
+  DIR *d;
+  char filename[MAXPATHLEN];
+  check_type_string(&x);
+  memcpy(filename,x->st.st_self,x->st.st_fillp);
+  filename[x->st.st_fillp]=0;
+  d=opendir(filename);
+  return (fixnum)d;
+}
+  
+DEFUN_NEW("D-TYPE-LIST",object,fSd_type_list,SI,0,0,NONE,OI,OO,OO,OO,(void),"") {
+  RETURN1(list(8,
+	       MMcons(make_fixnum(DT_BLK),make_keyword("BLOCK")),
+	       MMcons(make_fixnum(DT_CHR),make_keyword("CHAR")),
+	       MMcons(make_fixnum(DT_DIR),make_keyword("DIRECTORY")),
+	       MMcons(make_fixnum(DT_FIFO),make_keyword("FIFO")),
+	       MMcons(make_fixnum(DT_LNK),make_keyword("LINK")),
+	       MMcons(make_fixnum(DT_REG),make_keyword("FILE")),
+	       MMcons(make_fixnum(DT_SOCK),make_keyword("SOCKET")),
+	       MMcons(make_fixnum(DT_UNKNOWN),make_keyword("UNKNOWN"))
+	       ));
+}
+
+DEFUN_NEW("READDIR",object,fSreaddir,SI,2,2,NONE,OI,IO,OO,OO,(fixnum x,fixnum y),"") {
+  struct dirent *e;
+  object z;
+  if (!x) RETURN1(Cnil);
+  e=readdir((DIR *)x);
+  RETURN1(e ? make_simple_string(e->d_name) : Cnil);
+  for (;(e=readdir((DIR *)x)) && y!=DT_UNKNOWN && e->d_type!=y;);
+  if (!e) RETURN1(Cnil);
+  z=make_simple_string(e->d_name);
+  if (y==DT_UNKNOWN) z=MMcons(z,make_fixnum(e->d_type));
+  RETURN1(z);
+}
+
+DEFUN_NEW("CLOSEDIR",object,fSclosedir,SI,1,1,NONE,OI,OO,OO,OO,(fixnum x),"") {
+  closedir((DIR *)x);
+  return Cnil;
+}
+
+DEFUN_NEW("MKDIR",object,fSmkdir,SI,1,1,NONE,OO,OO,OO,OO,(object x),"") {
+
+  char filename[MAXPATHLEN];
+
+  check_type_string(&x);
+
+  memcpy(filename,x->st.st_self,x->st.st_fillp);
+  filename[x->st.st_fillp]=0;
+
+#ifdef __MINGW32__
+  if (mkdir(filename) < 0)
+#else        
+  if (mkdir(filename,01777) < 0)
+#endif        
+    FEerror("Cannot make the directory ~S.", 1, vs_base[0]);
+
+  RETURN1(x);
+
+}
 
 
 
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-#endif
-
 static void
 FFN(siLchdir)(void)
 {
